# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import ast
import copy
import enum
import logging
import typing
import warnings
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)

from gevent import (
    Timeout,
    event,
)
from pydantic.v1 import (
    Field,
    create_model,
)
from typing_extensions import (
    Literal,
    Self,
)
from types import MappingProxyType
from mxcubecore.CommandContainer import CommandContainer
from mxcubecore.dispatcher import dispatcher

if TYPE_CHECKING:
    from logging import Logger

    from pydantic.v1 import BaseModel


__copyright__ = """ Copyright © 2010-2020 by the MXCuBE collaboration """
__license__ = "LGPLv3+"


@enum.unique
class HardwareObjectState(enum.Enum):
    """Enumeration of common states, shared between all HardwareObjects"""

    UNKNOWN = 0
    WARNING = 1
    BUSY = 2
    READY = 3
    FAULT = 4
    OFF = 5


class DefaultSpecificState(enum.Enum):
    """Placeholder enumeration for HardwareObject-specific states"""

    UNKNOWN = "UNKNOWN"


class ConfiguredObject:
    """Superclass for classes that take configuration from YAML files"""

    # class HOConfig(pydantic.BaseModel):
    #     model_config = pydantic.ConfigDict(extra="allow")

    class HOConfig:
        """Temporary replacement for Pydantic class

        Required during transition, as long as we don't have the fields defined"""

        def __init__(self, **kwargs):
            self.__dict__.update(kwargs)

        def model_dump(self):
            return copy.deepcopy(self.__dict__)

    def __init__(
        self, name: str, hwobj_container: Optional["ConfiguredObject"] = None
    ) -> None:
        """
        Args:
            name (str): Equal to role name relative to hwobj_container (if applicable)
        """

        self._name = name
        self._config: Optional["ConfiguredObject.HOConfig"] = None
        self._hwobj_container: Optional[ConfiguredObject] = hwobj_container
        self._hwobj_by_role = {}

    @property
    def name(self):
        """HWOBJ name - Equal to role name relative to hwobj_container (if applicable)"""
        return self._name

    @property
    def id(self):
        """dot-separated role names defining path from beamline object to here

        NB beamline.id == ''"""
        if self._hwobj_container:
            names = []
            obj = self
            while obj._hwobj_container:
                names.append(obj.name)
                obj = obj._hwobj_container
            return ".".join(reversed(names))
        else:
            return ""

    @property
    def config(self):
        """Pydantic object with conifigured parameters, incl. contained HardwareObjects"""
        return self._config

    @property
    def hwobj_container(self):
        """HardwareObject contaiing this one - None for Beamline"""
        return self._hwobj_container

    def get_by_id(self, _id: str) -> "ConfiguredObject":
        result = self
        for name in _id.split("."):
            # Roles are no longer added to HardwareObjects as attributes so
            # we have to retreive them by role
            result = result.get_object_by_role(name)

            if not result:
                result = getattr(result, name)

            if result is None:
                break

        return result

    @property
    def objects_by_role(self) -> Dict[str, Union[Self, None]]:
        """All contained Objects mapped by role (in specification order).

        Includes objects defined in subclasses.

        Returns:
            OrderedDict[str, Union[Self, None]]: Contained objects mapped by role.
        """
        return MappingProxyType(self._hwobj_by_role)

    def get_properties(self) -> Dict[str, Any]:
        """Get configured properties (not roles)"""
        if self._config is not None:
            return self._config.model_dump()
        elif isinstance(self, HardwareObjectNode):
            # NBNB TEMPORARY for transition to yaml configuration only
            return HardwareObjectNode.get_properties(self)

    def get_property(self, name: str, default_value: Optional[Any] = None) -> Any:
        """Get property value .

        Args:
            name (str): Name
            default_value (Optional[Any], optional): Default value. Defaults to None.

        Returns:
            Any: Property value.
        """
        if self._config is not None:
            if hasattr(self._config, name):
                return getattr(self._config, name)
            else:
                return default_value
        elif isinstance(self, HardwareObjectNode):
            # NBNB TEMPORARY for transition to yaml configuration only
            return HardwareObjectNode.get_property(self, name, default_value)

    def get_roles(self) -> List[str]:
        """Get hardware object roles.

        Returns:
            List[str]: List of hardware object roles.
        """
        return list(self._hwobj_by_role.keys())

    def print_log(
        self,
        log_type: str = "HWR",
        level: str = "debug",
        msg: str = "",
    ) -> None:
        """Print message to logger.

        Args:
            log_type (str, optional): Logger type. Defaults to "HWR".
            level (str, optional): Logger level. Defaults to "debug".
            msg (str, optional): Message to log. Defaults to "".
        """
        if hasattr(logging.getLogger(log_type), level):
            getattr(logging.getLogger(log_type), level)(msg)


class PropertySet(dict):
    """Property Set"""

    def __init__(self) -> None:
        super().__init__()

        self.__properties_changed: Dict[str, Any] = {}
        self.__properties_path: Dict[str, Any] = {}

    def set_property_path(self, name: str, path: Union[str, Any]) -> None:
        """Set Property Path.

        Args:
            name (str): Name.
            path (Union[str, Any]): Path.
        """
        name = str(name)
        self.__properties_path[name] = path

    def get_properties_path(self) -> Iterator[tuple]:
        """Get Property Paths.

        Returns:
            Iterator[tuple]: Iterator for property paths.
        """
        return iter(self.__properties_path.items())

    def __setitem__(self, name: Union[str, Any], value: Union[str, Any]) -> None:
        name = str(name)

        if name in self and str(value) != str(self[name]):
            self.__properties_changed[name] = str(value)

        super().__setitem__(str(name), value)

    def get_changes(self) -> Generator[tuple, None, None]:
        """Get property changes since the last time checked.

        Yields:
            Generator[tuple, None, None]: Property changes.
        """
        for property_name, value in self.__properties_changed.items():
            yield (self.__properties_path[property_name], value)

        self.__properties_changed = {}  # reset changes at commit


class HardwareObjectNode:
    """Hardware Object Node"""

    user_file_directory: str

    def __init__(self, node_name: str) -> None:
        """
        Args:
            node_name (str): Node name.
        """
        self.__dict__["_property_set"] = PropertySet()
        self._property_set: PropertySet
        self.__objects_names: List[Union[str, None]] = []
        self.__objects: List[List[Union["HardwareObject", None]]] = []
        self._objects_by_role: Dict[str, "HardwareObject"] = {}
        self._path: str = ""
        self.__name: str = node_name
        self.__references: List[Tuple[str, str, str, int, int, int]] = []
        self._xml_path: Union[str, None] = None

    @staticmethod
    def set_user_file_directory(user_file_directory: str) -> None:
        """Set user file directory.

        Args:
            user_file_directory (str): User file directory path.
        """
        HardwareObjectNode.user_file_directory = user_file_directory

    @property
    def load_name(self) -> str:
        """Get node name.

        Returns:
            str: Name.
        """
        return self.__name

    def set_path(self, path: str) -> None:
        """Set the 'path' of the Hardware Object in the XML file describing it
        (the path follows the XPath syntax)

        Args:
            path (str): String representing the path of the Hardware Object in its file
        """

        # NB For use ONLY in loading xml-configured objects

        self._path = path

    def __iter__(self) -> Generator[Union["HardwareObject", None], None, None]:
        warnings.warn("%s.__iter__ is Deprecated. Avoid" % self.__class__.__name__)
        for i in range(len(self.__objects_names)):
            for object in self.__objects[i]:
                yield object

    def __len__(self) -> int:
        warnings.warn("%s.__len__ is Deprecated. Avoid" % self.__class__.__name__)
        return sum(map(len, self.__objects))

    def __getitem__(
        self,
        key: Union[str, int],
    ) -> Union["HardwareObject", List[Union["HardwareObject", None]], None]:
        warnings.warn("%s.__getitem__ is Deprecated. Avoid" % self.__class__.__name__)
        if isinstance(key, str):
            object_name = key

            try:
                index = self.__objects_names.index(object_name)
            except Exception:
                raise KeyError
            else:
                obj = self.__objects[index]
                if len(obj) == 1:
                    return obj[0]
                else:
                    return obj
        elif isinstance(key, int):
            index = key

            if index < len(self.__objects_names):
                obj = self.__objects[index]
                if len(obj) == 1:
                    return obj[0]
                else:
                    return obj
            else:
                raise IndexError
        else:
            raise TypeError

    def add_reference(
        self,
        name: str,
        reference: str,
        role: Union[str, None] = None,
    ) -> None:
        """Add hardware object reference.

        Args:
            name (str): Name.
            reference (str): Xpath reference.
            role (Union[str, None], optional): Role. Defaults to None.
        """

        role = str(role).lower()

        try:
            index = self.__objects_names.index(name)
        except ValueError:
            objects_names_index = len(self.__objects_names)
            self.__objects_names.append(None)
            objects_index = len(self.__objects)
            self.__objects.append(None)
            objects_index2 = -1
        else:
            objects_names_index = -1
            objects_index = index
            objects_index2 = len(self.__objects[index])
            self.__objects[index].append(None)

        self.__references.append(
            (reference, name, role, objects_names_index, objects_index, objects_index2)
        )

    def resolve_references(self) -> None:
        """Resolve hardware objects from defined references."""
        # NB Must be here - importing at top level leads to circular imports
        from .HardwareRepository import get_hardware_repository

        # NB For use ONLY in loading xml-configured objects

        while len(self.__references) > 0:
            (
                reference,
                name,
                role,
                objects_names_index,
                objects_index,
                objects_index2,
            ) = self.__references.pop()

            hw_object = get_hardware_repository().get_hardware_object(reference)

            if hw_object is not None:
                self._objects_by_role[role] = hw_object
                hw_object.__role = role

                if objects_names_index >= 0:
                    self.__objects_names[objects_names_index] = role
                    self.__objects[objects_index] = [hw_object]
                else:
                    self.__objects[objects_index][objects_index2] = hw_object
            else:
                if objects_names_index >= 0:
                    del self.__objects_names[objects_names_index]
                    del self.__objects[objects_index]
                else:
                    del self.__objects[objects_index][objects_index2]
                    if len(self.__objects[objects_index]) == 0:
                        del self.__objects[objects_index]

        # Inlined version of 'for hw_object in self' to avoid warnings
        # NB this function is used ONLY for loading XML-configured objects
        # and so will disappear
        for i in range(len(self.__objects_names)):
            for hw_object in self.__objects[i]:
                hw_object.resolve_references()

    def _add_object(
        self,
        name: str,
        hw_object: Union["HardwareObject", None],
        role: Optional[str] = None,
    ) -> None:
        """Add hardware object mapped by name.

        Args:
            name (str): Name.
            hw_object (Union[HardwareObject, None]): Hardware object.
            role (Optional[str], optional): Role. Defaults to None.
        """

        # NB For use ONLY in loading xml-configured objects

        if hw_object is None:
            return None
        elif role is not None:
            role = str(role).lower()
            self._objects_by_role[role] = hw_object
            hw_object.__role = role

        try:
            index = self.__objects_names.index(name)
        except ValueError:
            self.__objects_names.append(name)
            self.__objects.append([hw_object])
        else:
            self.__objects[index].append(hw_object)

    def _get_objects(
        self,
        object_name: str,
    ) -> Generator[Union["HardwareObject", None], None, None]:
        """Get hardware objects by name.

        Args:
            object_name (str): Name.

        Yields:
            Union[HardwareObject, None]: Hardware object.
        """

        # NB For use ONLY in loading xml-configured objects

        try:
            index = self.__objects_names.index(object_name)
        except ValueError:
            pass
        else:
            for obj in self.__objects[index]:
                yield obj

    def get_object_by_role(self, role: str) -> Union["HardwareObject", None]:
        """Get hardware object by role.

        Args:
            role (str): Role.

        Returns:
            Union[HardwareObject, None]: Hardware object.
        """
        role = str(role).lower()
        #
        # A hack to emulate get_object_by_role() for objects loaded from YAML config
        # files.
        #
        # When HWOBJ is loaded from YAML, we don't populate it's '_objects_by_role'
        # dictionary, thus that normal code path to look-up and object by role does
        # not work.
        #
        # However, objects are attached to the parents _hwobj_by_role object via
        # assignment. Try accessing using that attribute.
        #
        if hasattr(self, "_hwobj_by_role"):
            obj = self._hwobj_by_role.get(role, None)

            if obj is not None:
                return obj

        #
        # Look-up object by role the old way.
        #
        objects = [self]

        for curr in objects:
            result = curr._objects_by_role.get(role)
            if result is None:
                objects.extend(obj for obj in curr if obj)

            else:
                return result

    def _objects_names(self) -> List[Union[str, None]]:
        """Return hardware object names.

        Returns:
            List[Union[str, None]]: List of hardware object names.
        """
        return self.__objects_names[:]

    def _set_property(self, name: str, value: Any) -> None:
        """Set property value.

        Args:
            name (str): Name.
            value (Any): Value.
        """

        # NB For use ONLY in loading xml-configured objects

        name = str(name)
        value = str(value)

        if value == "None":
            value = None
        else:
            #
            # try to convert buffer to the appropriate type
            #
            try:
                value = int(value)
            except Exception:
                try:
                    value = float(value)
                except Exception:
                    if value == "True":
                        value = True
                    elif value == "False":
                        value = False

        self._property_set[name] = value
        self._property_set.set_property_path(name, self._path + "/" + str(name))

    def get_property(self, name: str, default_value: Optional[Any] = None) -> Any:
        """Get property value.

        Args:
            name (str): Name
            default_value (Optional[Any], optional): Default value. Defaults to None.

        Returns:
            Any: Property value.
        """
        return self._property_set.get(str(name), default_value)

    def get_properties(self) -> PropertySet:
        """Get properties - for XML-config implementation loading only

        Returns:
            PropertySet: Properties.
        """

        # NB For use ONLY in loading xml-configured objects

        return self._property_set


class HardwareObjectMixin(CommandContainer):
    """Functionality for either xml- or yaml-configured HardwareObjects

    Signals emited:

        - stateChanged

        - specificStateChanged
    """

    #: enum.Enum: General states, shared between all HardwareObjects. Do *not* overridde
    STATES = HardwareObjectState

    #: enum.Enum: Placeholder for HardwareObject-specific states. To be overridden
    SPECIFIC_STATES = DefaultSpecificState

    def __init__(self) -> None:
        CommandContainer.__init__(self)

        # Container for connections to HardwareObject
        self.connect_dict: Dict[str, Dict[str, Any]] = {}
        # event to handle waiting for object to be ready
        self._ready_event: event.Event = event.Event()
        # Internal general state attribute, used to check for state changes
        self._state: Union[HardwareObjectState, None] = None
        # Internal object-specific state attribute, used to check for state changes
        self._specific_state: Union[Any, None] = None

        # Dictionary on the form:
        # key: The exporterd member function name
        # value: The arguments of the exported member
        self._exports: Dict[str, Any] = {}

        # Dictionary containing list Pydantic models for each of the exported member
        # functions arguemnts. The key is the member name and the value a list of the
        # pydantic models.
        self._pydantic_models: Dict[str, Type["BaseModel"]] = {}

        # Dictionary on the form:
        # key: The exporterd member function name
        # value: dictionary on the form {signautre: [<arguments>] schema:<JSONSchema>}
        self._exported_attributes: Dict[str, Any] = {}

        # List of member names (methods) to be exported (Set at configuration stage)
        self._exports_config_list = []

        self.log: "Logger" = logging.getLogger("HWR").getChild(self.__class__.__name__)
        self.user_log: "Logger" = logging.getLogger("user_log_level")

    def __bool__(self) -> Literal[True]:
        return True

    def __nonzero__(self) -> Literal[True]:
        return True

    def _init(self) -> None:
        """'protected' post-initialization method. Override as needed.

        For ConfiguredObjects called before loading contained objects
        """
        self.update_state(self.STATES.UNKNOWN)

    def init(self) -> None:
        """'public' post-initialization method. Override as needed.

        For ConfiguredObjects called after loading contained objects.
        """
        self._exports_config_list.extend(
            ast.literal_eval(self.get_property("exports", "[]").strip())
        )
        self._exports = dict.fromkeys(self._exports_config_list, {})

        # Add methods that are exported programatically
        for attr_name in dir(self):
            _attr = getattr(self, attr_name)
            if getattr(_attr, "__exported__", False):
                self._exports[attr_name] = []

        if self._exports:
            self._get_type_annotations()

    def _get_type_annotations(self) -> None:
        """Retrieve typehints and create pydantic models for each argument."""
        _models = {}

        for attr_name, _ in self._exports.items():
            self._exported_attributes[attr_name] = {}
            self._exports[attr_name] = []
            self._pydantic_models[attr_name] = {}
            fdict = {}

            try:
                _attr = getattr(self, attr_name)
            except AttributeError:
                logging.getLogger("HWR").error(
                    f"{attr_name} configured as exported for {self.name} but not implemented"
                )
                continue

            for _n, _t in typing.get_type_hints(_attr).items():
                # Skipp return typehint
                if _n != "return":
                    self._exports[attr_name].append(_n)
                    fdict[_n] = (_t, Field(alias=_n))

            _models[attr_name] = (
                create_model(attr_name, **fdict),
                Field(alias=attr_name),
            )

            self._pydantic_models[attr_name] = _models[attr_name][0]
            self._exported_attributes[attr_name]["display"] = True
            self._exported_attributes[attr_name]["signature"] = self._exports[attr_name]
            self._exported_attributes[attr_name]["schema"] = self._pydantic_models[
                attr_name
            ].schema_json()

        model = create_model(self.__class__.__name__, **_models)
        self._pydantic_models["all"] = model

    def execute_exported_command(self, cmd_name: str, args: Dict[str, Any]) -> Any:
        """Execute exported command.

        Args:
            cmd_name (str): Command name.
            args (Dict[str, Any]): Command arguments.

        Returns:
            Any: Command execution output.
        """
        if cmd_name in self._exports.keys():
            cmd = getattr(self, cmd_name)
        else:
            self.log.info(
                f"Command {cmd_name} not exported, check type hints and configuration file"
            )

        return cmd(**args)

    @property
    def pydantic_model(self) -> Dict[str, Type["BaseModel"]]:
        """Get object Pydantic models.

        Returns:
            Dict[str, Type[BaseModel]]: Pydantic models for object.
        """
        return self._pydantic_models

    @property
    def exported_attributes(self) -> Dict[str, Any]:
        """Get exported attributes.

        Returns:
            Dict[str, Any]: Dictionary containing the method signature and JSONSchema.
            Follows the format;
            ```python
            {
                "schema": <JSONSchema string>,
                "signaure": <list of argument names>,
            }
            ```
        """
        return self._exported_attributes

    def abort(self) -> None:
        """Immediately terminate HardwareObject action.

        Should not happen in state READY.
        """
        if self.get_state() is self.STATES.READY:
            return

        # When overriding put active code here

    def stop(self) -> None:
        """Gentler (?) alternative to abort.

        Override as necessary to implement.
        """
        self.abort()

    def get_state(self) -> HardwareObjectState:
        """Getter for state attribute.

        Implementations must query the hardware directly, to ensure current results.

        Returns:
            HardwareObjectState: Current state.
        """
        return self._state

    def get_specific_state(self) -> Union[Any, None]:
        """Getter for specific_state attribute. Override if needed.

        Returns:
            Union[Any, None]: Specific state enum or None.
        """
        return self._specific_state

    def wait_ready(self, timeout: Optional[float] = None) -> None:
        """Wait timeout seconds till object is ready.

        If timeout is None: wait forever.

        Args:
            timeout (Optional[float], optional): Timeout (seconds). Defaults to None.
        """
        with Timeout(timeout, RuntimeError("Timeout waiting for status ready")):
            self._ready_event.wait(timeout=timeout)

    def is_ready(self) -> bool:
        """Convenience function: Check if the object state is READY.

        The same effect could be achieved with
        ```python
        self.get_state() == self.STATES.READY
        ```

        Returns:
            bool: True if ready, otherwise False.
        """
        return self._ready_event.is_set()

    def set_is_ready(self, value: bool):
        warnings.warn(
            "set_is_ready method ported from Device is Deprecated and will be removed",
            DeprecationWarning,
        )
        if value:
            self.update_state(HardwareObjectState.READY)
        else:
            self.update_state(HardwareObjectState.OFF)

    def update_state(self, state: Optional[HardwareObjectState] = None) -> None:
        """Update self._state, and emit signal stateChanged if the state has changed.

        Args:
            state (Optional[HardwareObjectState], optional): State. Defaults to None.

        Raises:
            ValueError: If state specified is invalid.
        """
        if state is None:
            state = self.get_state()

        is_set = self._ready_event.is_set()

        if state == self.STATES.READY:
            if not is_set:
                self._ready_event.set()
        elif not isinstance(state, self.STATES):
            raise ValueError("Attempt to update to illegal state: %s" % state)
        elif is_set:
            self._ready_event.clear()

        if state != self._state:
            self._state = state
            self.emit("stateChanged", (self._state,))

    def update_specific_state(self, state: Optional[Any] = None) -> None:
        """Update self._specific_state, and emit specificStateChanged if appropriate.

        Args:
            state (Optional[Any], optional): Specific state - the enumeration will be
            specific for each HardwareObject class. Defaults to None.

        Raises:
            ValueError: If state specified is invalid.
        """
        if state is None:
            state = self.get_specific_state()
        if state != self._specific_state:
            if not isinstance(state, self.SPECIFIC_STATES):
                raise ValueError(
                    "Attempt to update to illegal specific state: %s" % state
                )

            self._specific_state = state
            self.emit("specificStateChanged", (state,))

    def re_emit_values(self) -> None:
        """Update values for all internal attributes

        Should be expanded in subclasse with more updatable attributes
        (e.g. value, limits)
        """
        self.update_state()
        self.update_specific_state()

    def force_emit_signals(self) -> None:
        """Emits all hardware object signals

        The method is called from the gui via beamline object to ensure that bricks have values
        after the initialization.
        Problem arrise when a hardware object is used by several bricks.
        If first brick connects to some signal emited by a brick then
        other bricks connecting to the same signal will not receive the
        values on the startup.
        The easiest solution is to call force_emit_signals method directly
        after the initialization of the beamline object
        """
        pass

    # Moved from HardwareObjectNode
    def clear_gevent(self) -> None:
        """Clear gevent tasks, called when disconnecting a HardwareObject.

        Override in subclasses as needed.
        """
        self.update_state(self.STATES.UNKNOWN)

    # Signal handling functions:
    def emit(self, signal: Union[str, object, Any], *args) -> None:
        """Emit signal. Accepts both multiple args and a single tuple of args.

        TODO: This function would be unnecessary if all callers used
        ```python
        dispatcher.send(signal, self, *argtuple)
        ```

        Args:
            signal (Union[str, object, Any]): In practice a string, or dispatcher.
            *args (tuple): Arguments sent with signal.
        """

        signal = str(signal)

        if len(args) == 1:
            if isinstance(args[0], tuple):
                args = args[0]
        dispatcher.send(signal, self, *args)

    def connect(
        self,
        sender: Union[str, object, Any],
        signal: Union[str, Any],
        slot: Optional[Callable] = None,
    ) -> None:
        """Connect a signal sent by self to a slot.

        The functions provides syntactic sugar ; Instead of
        ```python
        self.connect(self, "signal", slot)
        ```
        it is possible to do
        ```python
        self.connect("signal", slot)
        ```

        TODO: This would be much nicer if refactored as
        ```python
        def connect(self, signal, slot, sender=None)
        ```

        Args:
            sender (Union[str, object, Any]): If a string, interprted as the signal.
            signal (Union[str, Any]): In practice a string, or dispatcher.
            Any if sender is a string interpreted as the slot.
            slot (Optional[Callable], optional): In practice a functon or method.
            Defaults to None.

        Raises:
            ValueError: If slot is None and "sender" parameter is not a string.
        """

        if slot is None:
            if isinstance(sender, str):
                slot = signal
                signal = sender
                sender = self
            else:
                raise ValueError("invalid slot (None)")

        signal = str(signal)

        dispatcher.connect(slot, signal, sender)

        self.connect_dict[sender] = {"signal": signal, "slot": slot}

        if hasattr(sender, "connect_notify"):
            sender.connect_notify(signal)

    def disconnect(
        self,
        sender: Union[str, object, Any],
        signal: Union[str, Any],
        slot: Optional[Callable] = None,
    ) -> None:
        """Disconnect a signal sent by self to a slot.

        The functions provides syntactic sugar ; Instead of
        ```python
        self.connect(self, "signal", slot)
        ```
        it is possible to do
        ```python
        self.connect("signal", slot)
        ```

        TODO: This would be much nicer if refactored as
        ```python
        def disconnect(self, signal, slot, sender=None): ...
        ```

        Args:
            sender (Union[str, object, Any]): If a string, interprted as the signal.
            signal (Union[str, Any]): In practice a string, or dispatcher.
            Any if sender is a string interpreted as the slot.
            slot (Optional[Callable], optional): In practice a functon or method.
            Defaults to None.

        Raises:
            ValueError: If slot is None and "sender" parameter is not a string.
        """
        if slot is None:
            if isinstance(sender, str):
                slot = signal
                signal = sender
                sender = self
            else:
                raise ValueError("invalid slot (None)")

        signal = str(signal)

        dispatcher.disconnect(slot, signal, sender)

        if hasattr(sender, "disconnect_notify"):
            sender.disconnect_notify(signal)

    # def connect_notify(self, signal):
    #     pass
    #
    # def disconnect_notify(self, signal):
    #     pass


class HardwareObject(ConfiguredObject, HardwareObjectNode, HardwareObjectMixin):
    """Xml-configured hardware object"""

    def __init__(self, name: str) -> None:
        """
        Args:
            name (str): Name.
        """
        ConfiguredObject.__init__(self, name)
        HardwareObjectNode.__init__(self, name)
        HardwareObjectMixin.__init__(self)

    def __getstate__(self) -> str:
        # NBNB Needs updating

        return self.load_name

    def __setstate__(self, name: str) -> None:
        # NB Must be here - importing at top level leads to circular imports
        from .HardwareRepository import get_hardware_repository

        # NBNB Needs updating

        obj = get_hardware_repository().get_hardware_object(name)
        self.__dict__.update(obj.__dict__)

    def commit_changes(self) -> None:
        """Commit last changes back to configuration."""

        # NB Must be here - importing at top level leads to circular imports
        from .HardwareRepository import get_hardware_repository

        def get_changes(node: Self) -> Union[list, List[tuple], Any]:
            updates = list(node._property_set.get_changes())
            if node:
                for subnode in node:
                    updates += get_changes(subnode)

            if isinstance(node, HardwareObject):
                if updates:
                    get_hardware_repository().update(node.load_name, updates)
                return []
            else:
                return updates

        get_changes(self)

    def rewrite_xml(self, xml: Union[bytes, Any]) -> None:
        """Rewrite XML conifguration file.

        Args:
            xml (Union[bytes, Any]): XML source to write to file.
        """

        # NB Must be here - importing at top level leads to circular imports
        from .HardwareRepository import get_hardware_repository

        get_hardware_repository().rewrite_xml(self.load_name, xml)

    def xml_source(self) -> Union[str, Any]:
        """Get XML configuration source.

        Returns:
            Union[str, Any]: XML source.
        """

        # NB Must be here - importing at top level leads to circular imports
        from .HardwareRepository import get_hardware_repository

        return get_hardware_repository().xml_source[self.load_name]


HardwareObjectYaml = HardwareObject


class Procedure(HardwareObject):
    """Procedure"""

    def __init__(self, name: str) -> None:
        """
        Args:
            name (str): Name.
        """
        HardwareObject.__init__(self, name)

    def userName(self) -> str:
        """Get procedure user name.

        Returns:
            str: Username.
        """
        uname = self.get_property("username")
        if uname is None:
            return str(self.name())
        else:
            return uname

    def GUI(self, parent: Any) -> None:
        pass


class Device(HardwareObject):
    """Old superclass for devices

    Signals:

        - "deviceReady"

        - "deviceNotReady"

    NB Deprecated - should be replaced by AbstractActuator"""

    (NOTREADY, READY) = (0, 1)  # device states

    def __init_subclass__(cls, **kwargs):
        warnings.warn(
            f"{cls.__name__} will be deprecated.", DeprecationWarning, stacklevel=2
        )
        super().__init_subclass__(**kwargs)

    def __init__(self, name):
        warnings.warn(
            "class Device is Deprecated and will be removed",
            DeprecationWarning,
        )
        HardwareObject.__init__(self, name)

        self.state = Device.NOTREADY

    def set_is_ready(self, ready):
        if ready and self.state == Device.NOTREADY:
            self.state = Device.READY
            self.emit("deviceReady")
        elif not ready and self.state == Device.READY:
            self.state = Device.NOTREADY
            self.emit("deviceNotReady")

    def is_ready(self):
        return self.state == Device.READY

    def userName(self):
        # TODO standardise on 'username' or 'user_name' globaly
        uname = self.get_property("username")
        if uname is None:
            return str(self.name())
        else:
            return uname


class DeviceContainer:
    """Device ocntainer class - old style

    NB Deprecated. Once DeviceContainerNode is removed,
    this clould be merged into Equipment"""

    def __init__(self):
        pass

    def get_devices(self):
        devices = []

        for item in dir(self):
            if isinstance(item, Device):
                devices.append(item)
            elif isinstance(item, DeviceContainer):
                devices += item.get_devices()

        return devices

    def get_device(self, device_name):
        devices = self.get_devices()

        for device in devices:
            if str(device.name()) == device_name:
                return device

    def get_device_by_role(self, role):
        # TODO This gives a pylint error, since get_object_by_roleis not in a superclass
        # it is available in the subclases that use this, but fixing this
        # would make more sense in connection with a general refactoring of
        # Device / DeciveContainer/Equipment
        item = self.get_object_by_role(role)

        if isinstance(item, Device):
            return item

    def clear_gevent(self):
        pass


class DeviceContainerNode(HardwareObjectNode, DeviceContainer):
    """Class serves solely to provide a single supercalss
    combining HardwareObjectNode and DeviceContaine

    TODO.This class is Deprecated.

    it is only used once,
    in HardwareObjectFileParser.HardwareObjectHandler.startElement
    And that use looks like it could be replaced by something else"""


class Equipment(HardwareObject, DeviceContainer):
    """Equipment class -old style

    Signals:

        - equipmentReady"

        - "equipmentNotReady"

    NB This class needs refactoring. Since many (soon: all??) contained
     objects are no longer of class Device, the code in here is unlikely to work."""

    def __init_subclass__(cls, **kwargs):
        warnings.warn(
            f"{cls.__name__} will be deprecated.", DeprecationWarning, stacklevel=2
        )
        super().__init_subclass__(**kwargs)

    def __init__(self, name):
        HardwareObject.__init__(self, name)
        DeviceContainer.__init__(self)
        warnings.warn(
            "class Equipment is Deprecated and will be removed",
            DeprecationWarning,
        )

        self.__ready = None

    def _init(self):
        for device in self.get_devices():
            self.connect(device, "deviceReady", self.__device_ready)
            self.connect(device, "deviceNotReady", self.__device_not_ready)

        self.__device_ready()

    def __device_ready(self):
        ready = True
        for device in self.get_devices():
            ready = ready and device.is_ready()
            if not ready:
                break

        if self.__ready != ready:
            self.__ready = ready

            if self.is_ready():
                self.emit("equipmentReady")
            else:
                self.emit("equipmentNotReady")

    def __device_not_ready(self):
        if self.__ready:
            self.__ready = False
            self.emit("equipmentNotReady")

    def is_ready(self):
        return self.is_valid() and self.__ready

    def is_valid(self):
        return True

    def userName(self):
        uname = self.get_property("username")
        if uname is None:
            return str(self.name())
        else:
            return uname


class Null:
    """
    This class ignores all parameters passed when constructing or
    calling instances and traps all attribute and method requests.
    Instances of it always (and reliably) do 'nothing'.

    The code might benefit from implementing some further special
    Python methods depending on the context in which its instances
    are used. Especially when comparing and coercing Null objects
    the respective methods' iimplementation will depend very much
    on the environment and, hence, these special methods are not
    provided here.
    """

    def __init__(self, *args, **kwargs):
        "Ignore parameters."
        return None

    def __call__(self, *args, **kwargs):
        "Ignore method calls."
        return self

    def __bool__(self):
        return 0

    def __getattr__(self, mname):
        "Ignore attribute requests."
        return self

    def __setattr__(self, name, value):
        "Ignore attribute setting."
        return self

    def __delattr__(self, name):
        "Ignore deleting attributes."
        return self

    def __repr__(self):
        "Return a string representation."
        return "<Null>"

    def __str__(self):
        "Convert to a string and return it."
        return "Null"



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import glob
import json
import os
import time
from configparser import ConfigParser
from datetime import date
from select import EPOLL_CLOEXEC

import yaml

from mxcubecore.HardwareObjects.Session import Session

PATH_BEAMTIME = "/gpfs/current"
PATH_COMMISSIONING = "/gpfs/commissioning"
PATH_FALLBACK = "/gpfs/local"


class P11Session(Session):
    default_archive_folder = "raw"

    def init(self):
        super().init()

        self.settings_file = self.get_property("p11_settings_file")
        self.operation_mode = self.get_property("mode")
        self.beamtime_info = {}

        parser = ConfigParser()
        parser.read(self.settings_file)
        self.session_file_name = parser["general"]["file_name"]
        self.session_file_name = "mxcube"

        if self.session_start_date:
            self.start_time = self.session_start_date.split(" ")[0].replace("-", "")
        else:
            self.start_time = time.strftime("%Y%m%d")

        self.info_set_defaults()

        # Try to locate the metadata file and process it.
        self.beamtime_metadata_file = self.locate_metadata_file()

        if self.beamtime_metadata_file:
            (
                self.beamline,
                self.beamtime,
                self.remote_data_dir,
                self.user_name,
                self.user_sshkey,
                self.slurm_reservation,
                self.slurm_partition,
                self.slurm_node,
            ) = self.parse_metadata_file(self.beamtime_metadata_file)
        else:
            # Fall back to local paths if no metadata is found
            self.log.debug("Falling back to local directory for saving data.")
            self.select_base_directory(
                "local"
            )  # Use local paths instead of beamtime data

        if self.is_beamtime_open():
            self.read_beamtime_info()
        elif self.is_commissioning_open():
            self.read_commissioning_info()

        self.set_base_data_directories(
            self.base_directory,
            self.base_directory,
            self.base_directory,
            raw_folder=self.raw_data_folder_name,
            process_folder=self.processed_data_folder_name,
            archive_folder=self.default_archive_folder,
        )

    def info_set_defaults(self):
        self.beamtime_info["beamtimeId"] = None
        self.beamtime_info["proposalType"] = None
        self.beamtime_info["proposalId"] = None
        self.beamtime_info["rootPath"] = PATH_FALLBACK

    def is_beamtime_open(self):
        self.log.debug("=========== CHECKING IF BEAMTIME ID IS OPEN... ============")
        if self.is_writable_dir(os.path.join(PATH_BEAMTIME, self.raw_data_folder_name)):
            self.log.debug(
                "=========== BEAMTIME IS OPEN (/gpfs/current exists) ============"
            )
        else:
            logging.getLogger("GUI").error(
                "No beamtime ID is open (check /gpfs/current)"
            )

        return self.is_writable_dir(
            os.path.join(PATH_BEAMTIME, self.raw_data_folder_name)
        )

    def is_commissioning_open(self):
        return self.is_writable_dir(
            os.path.join(PATH_COMMISSIONING, self.raw_data_folder_name)
        )

    def is_writable_dir(self, folder):
        return os.path.isdir(folder) and os.access(folder, os.F_OK | os.W_OK)

    def get_current_beamtime_id(self):
        if self.is_beamtime_open():
            info = self.get_beamtime_info()
            return info["beamtimeId"]

    def get_current_proposal_code(self):
        if self.is_beamtime_open():
            info = self.get_beamtime_info()
            return info["proposalType"]

    def get_current_proposal_number(self):
        if self.is_beamtime_open():
            info = self.get_beamtime_info()
            return info["proposalId"]

    def get_beamtime_info(self):
        return self.beamtime_info

    def read_beamtime_info(self):
        self.log.debug("=========== READING BEAMTIME INFO ============")
        if os.path.exists(PATH_BEAMTIME):
            if os.scandir(PATH_BEAMTIME):
                for ety in os.scandir(PATH_BEAMTIME):
                    if ety.is_file() and ety.name.startswith("beamtime-metadata"):
                        info = self.read_load_info(ety.path)
                        self.log.debug(f"BEAMTIME INFO from {ety.path} is " + str(info))
                        if info is not None:
                            self.beamtime_info.update(self.read_load_info(ety.path))
                        self.beamtime_info["rootPath"] = PATH_BEAMTIME
        else:
            self.log.debug(f"No beamtime ID is open, using local path {PATH_FALLBACK}.")
            self.beamtime_info["rootPath"] = PATH_FALLBACK

    def read_commissioning_info(self):
        for ety in os.scandir(PATH_COMMISSIONING):
            if ety.is_file() and ety.name.startswith("commissioning-metadata"):
                fname = ety.path
                break
            else:
                return None

        return self.read_load_info(fname)

    def read_load_info(self, filename):
        try:
            with open(filename, encoding="utf-8") as file:
                json_str = file.read()
                return json.loads(json_str)
        except (ValueError, FileNotFoundError):
            return None

    def select_base_directory(self, mode="beamtime"):
        self.base_directory = self.beamtime_info["rootPath"]

    def get_base_data_directory(self):
        """
        Returns the base data directory taking the 'contextual'
        information into account, such as if the current user
        is inhouse.

        :returns: The base data path.
        :rtype: str
        """
        user_category = ""
        directory = ""

        return self.base_directory

    def get_base_image_directory(self):
        """
        :returns: The base path for images.
        :rtype: str
        """
        return os.path.join(self.get_base_data_directory(), self.raw_data_folder_name)

    def get_base_process_directory(self):
        """
        :returns: The base path for procesed data.
        :rtype: str
        """
        return os.path.join(
            self.get_base_data_directory(), self.processed_data_folder_name
        )

    def get_archive_directory(self):
        """
        :returns: The base path for procesed data.
        :rtype: str
        """
        return os.path.join(self.get_base_data_directory(), self.default_archive_folder)

    def path_to_ispyb(self, path):
        ispyb_template = self["file_info"].get_property("ispyb_directory_template")
        bid = self.beamtime_info["beamtimeId"]
        year = date.today().year
        ispyb_path = ispyb_template.format(beamtime_id=bid, year=year)
        return path

    def is_writable_dir(self, folder):
        return os.path.isdir(folder) and os.access(folder, os.F_OK | os.W_OK)

    def locate_metadata_file(self, root_dir="/gpfs"):
        try:
            beamtime_dirs = [
                path
                for path in [
                    os.path.join(root_dir, entry) for entry in os.listdir(root_dir)
                ]
                if os.path.isdir(path) and not path.endswith("local")
            ]
        except OSError as e:
            print(e)
            self.log.debug("Root directory does not exist: " + str(root_dir))
            return None  # Fall back if the root directory doesn't exist.

        self.log.debug(f"Scanning directories: {beamtime_dirs}")

        metadata_files = []
        for curr_dir in beamtime_dirs + [root_dir]:
            curr_dir_metadata_files = glob.glob("{0}/*metadata*.json".format(curr_dir))
            metadata_files.extend(curr_dir_metadata_files)
            self.log.debug(
                f"Found metadata files in {curr_dir}: {curr_dir_metadata_files}"
            )

        if len(metadata_files) != 1:
            self.log.debug(
                "Unique metadata JSON file not found. Falling back to /gpfs/local."
            )
            return None  # Return None to indicate no metadata file was found.

        return metadata_files[0]

    def parse_metadata_file(self, metadatafile_path):
        beamline = ""
        beamtime = ""
        coredatadir = ""
        temp_user_name = ""
        temp_user_sshkeyfile = ""
        slurm_reservation = ""
        slurm_partition = ""
        reserved_nodes = []
        with open(metadatafile_path, "r") as mdfile:
            try:
                md = yaml.safe_load(mdfile)
                if "beamline" in md:
                    beamline = str(md["beamline"])
                if "beamtimeId" in md:
                    beamtime = str(md["beamtimeId"])
                elif "id" in md:
                    beamtime = str(md["id"])
                if "corePath" in md:
                    coredatadir = str(md["corePath"])
                if "onlineAnalysis" in md:
                    temp_user_name = str(md["onlineAnalysis"]["userAccount"])
                    temp_user_sshkeyfile = str(
                        md["onlineAnalysis"]["sshPrivateKeyPath"]
                    )
                    slurm_reservation = str(md["onlineAnalysis"]["slurmReservation"])
                    slurm_partition = str(md["onlineAnalysis"]["slurmPartition"])
                    reserved_nodes = md["onlineAnalysis"]["reservedNodes"]
            except:
                raise RuntimeError(
                    "JSON parsing of metadata file failed", metadatafile_path
                )

        if not beamline:
            raise RuntimeError("Beamline ID not found", metadatafile_path)
        if not beamtime:
            raise RuntimeError("Beamtime ID not found", metadatafile_path)
        if not coredatadir:
            raise RuntimeError(
                "Data location on remote filesystem unknown", metadatafile_path
            )
        if not temp_user_name:
            raise RuntimeError(
                "Temporary account for online analysis unknown ", metadatafile_path
            )
        if not temp_user_sshkeyfile:
            raise RuntimeError(
                "SSH key for online analysis account not found", metadatafile_path
            )
        if not slurm_reservation:
            raise RuntimeError(
                "Slurm reservation for online analysis not found", metadatafile_path
            )
        if not slurm_partition:
            raise RuntimeError(
                "Slurm partition for online analysis not found", metadatafile_path
            )
        if not reserved_nodes:
            raise RuntimeError(
                "Reserved node(s) for online analysis not found", metadatafile_path
            )
        else:
            temp_user_sshkeyfile = os.path.join(
                os.path.dirname(metadatafile_path), temp_user_sshkeyfile
            )
            slurm_node = str(reserved_nodes[0])

        return (
            beamline,
            beamtime,
            coredatadir,
            temp_user_name,
            temp_user_sshkeyfile,
            slurm_reservation,
            slurm_partition,
            slurm_node,
        )

    def get_beamtime_metadata(self, root_dir="/gpfs"):
        try:
            metadata_file = self.locate_metadata_file(root_dir)
            return self.parse_metadata_file(metadata_file)
        except:
            self.log.debug(
                "Metadata file can not be parsed. Check if beamtime is open."
            )
            return None

    def get_ssh_command(self):
        ssh_command = "/usr/bin/ssh"
        ssh_opts_general = "-o BatchMode=yes -o CheckHostIP=no -o StrictHostKeyChecking=no -o GSSAPIAuthentication=no -o GSSAPIDelegateCredentials=no -o PasswordAuthentication=no -o PubkeyAuthentication=yes -o PreferredAuthentications=publickey -o ConnectTimeout=10"
        ssh_opts_user = "-l {0}".format(self.user_name)
        ssh_opts_key = "-i {0}".format(self.user_sshkey)
        ssh_opts_host = self.slurm_node
        ssh_command += " {0} {1} {2} {3}".format(
            ssh_opts_general, ssh_opts_key, ssh_opts_user, ssh_opts_host
        )
        return ssh_command

    def get_sbatch_command(
        self, jobname_prefix="onlineanalysis", logfile_path="/dev/null"
    ):
        sbatch_command = "/usr/bin/sbatch"
        sbatch_opts_jobname = "{0}_{1.beamline}_{1.beamtime}".format(
            jobname_prefix, self
        )

        sbatch_opts_logfile = logfile_path

        sbatch_opts = "--partition={0.slurm_partition} --reservation={0.slurm_reservation} --job-name={1} --output={2}".format(
            self, sbatch_opts_jobname, sbatch_opts_logfile
        )

        sbatch_command += " {0}".format(sbatch_opts)
        return sbatch_command



try:
    from louie import dispatcher  # pyright: ignore[reportMissingImports]
    from louie import robustapply  # pyright: ignore[reportMissingImports]
    from louie import saferef  # pyright: ignore[reportMissingImports]

    louie = 1
except ImportError:
    from pydispatch import dispatcher  # pyright: ignore[reportMissingImports]
    from pydispatch import robustapply  # pyright: ignore[reportMissingImports]
    from pydispatch import saferef  # pyright: ignore[reportMissingImports]

    saferef.safe_ref = saferef.safeRef
    robustapply.robust_apply = robustapply.robustApply
    louie = 0

import sys

if not hasattr(robustapply, "_robust_apply"):
    # patch robustapply.robust_apply to display exceptions, but to ignore them
    # this makes 'dispatcher.send' to continue on exceptions, which is
    # the behaviour we want ; it's not because a receiver doesn't handle a
    # signal properly that the whole chain should stop
    robustapply._robust_apply = robustapply.robust_apply

    def __my_robust_apply(*args, **kwargs):
        try:
            return robustapply._robust_apply(*args, **kwargs)
        except Exception:
            sys.excepthook(*sys.exc_info())

    if louie:
        robustapply.robust_apply = __my_robust_apply
    else:
        robustapply.robustApply = __my_robust_apply
    del louie
    del __my_robust_apply



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

from enum import Enum

from mxcubecore.BaseHardwareObjects import HardwareObjectState
from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractNState

__credits__ = ["DESY P11"]
__license__ = "LGPLv3+"
__category__ = "General"

from enum import Enum

from mxcubecore.HardwareObjects.abstract.AbstractShutter import AbstractShutter


class P11FastShutter(AbstractNState):
    """
    P11 BakcLight define interface to Tango backlight at DESY P11
    """

    default_open_time = 8
    default_close_time = 3

    def __init__(self, name):
        super().__init__(name)
        self.chan_value = None

    def init(self):
        """Initilise the predefined values"""

        self._initialise_values()
        self.chan_value = self.get_channel_object("value")

        if self.chan_value is not None:
            self.chan_value.connect_signal("update", self.update_fast_shutter)

        self.update_fast_shutter(self.chan_value.get_value())
        super().init()

    def _initialise_values(self):
        """Add additional, known in advance states to VALUES"""
        values_dict = {item.name: item.value for item in self.VALUES}
        values_dict.update({"OPEN": "Open", "CLOSED": "Closed"})
        self.VALUES = Enum("ValueEnum", values_dict)

    def get_value(self):
        return self.update_fast_shutter()

    @property
    def is_open(self):
        return self.get_value() == self.VALUES.OPEN

    @property
    def is_closed(self):
        return self.get_value() == self.VALUES.CLOSED

    def open(self):
        self.set_value(self.VALUES.OPEN)

    def close(self):
        self.set_value(self.VALUES.CLOSED)

    def _set_value(self, value):
        current_value = self.chan_value.get_value()

        if value == self.VALUES.OPEN:
            new_value = 1
        elif value == self.VALUES.CLOSED:
            new_value = 0

        if current_value != new_value:
            self.chan_value.set_value(new_value)

    def update_fast_shutter(self, value=None):
        """Updates shutter state

        :return: shutter state as str
        """
        if value is None:
            value = self.chan_value.get_value()

        if value:
            shutter_value = self.VALUES.OPEN
        else:
            shutter_value = self.VALUES.CLOSED

        self.update_value(shutter_value)

        return shutter_value



#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__credits__ = ["EMBL Hamburg"]
__license__ = "LGPLv3+"
__category__ = "General"


class EMBLDoorInterlock(HardwareObject):

    DoorInterlockState = {
        3: "unlocked",
        1: "closed",
        0: "locked_active",
        46: "locked_inactive",
        -1: "error",
    }

    def __init__(self, name):

        super().__init__(name)

        self.use_door_interlock = None
        self.door_interlock_state = None
        self.door_interlock_final_state = None
        self.door_interlock_breakabled = None

        self.chan_ics_error = None
        self.chan_state_locked = None
        self.chan_state_breakable = None
        self.chan_cmd_break_error = None
        self.cmd_break_interlock = None

        self.ics_enabled = True

    def init(self):

        self.door_interlock_state = "unknown"

        self.use_door_interlock = self.get_property("useDoorInterlock", True)

        self.chan_state_locked = self.get_channel_object("chanStateLocked")
        self.chan_state_locked.connect_signal("update", self.state_locked_changed)
        self.chan_state_breakable = self.get_channel_object("chanStateBreakable")
        self.chan_state_breakable.connect_signal("update", self.state_breakable_changed)

        self.chan_ics_error = self.get_channel_object("chanIcsErrorOne")
        self.chan_ics_error.connect_signal("update", self.ics_error_msg_changed)

        self.chan_cmd_break_error = self.get_channel_object("chanCmdBreakError")
        if self.chan_cmd_break_error is not None:
            self.chan_cmd_break_error.connect_signal(
                "update", self.cmd_break_error_msg_changed
            )

        self.cmd_break_interlock = self.get_command_object("cmdBreak")

    def cmd_break_error_msg_changed(self, error_msg):
        """Displays error log message if door interlock break do not work

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0:
            logging.getLogger("GUI").error(
                "Break ICS door interlock: Error %s" % error_msg
            )

    def ics_error_msg_changed(self, error_msg):
        """Updates ICS error message

        :param error_msg: error message
        :type error_msg: str
        :return: None
        """
        if len(error_msg) > 0:
            self.ics_enabled = False
        else:
            self.ics_enabled = True
        self.get_state()

    def connected(self):
        """Sets is ready"""
        self.set_is_ready(True)

    def disconnected(self):
        """Sets not ready"""
        self.set_is_ready(False)

    def state_breakable_changed(self, state):
        """Updates door interlock state"""
        self.door_interlock_breakabled = state
        self.get_state()

    def state_locked_changed(self, state):
        """Updates door interlock state"""
        self.door_interlock_state = state
        self.get_state()

    def get_state(self):
        """Returns current state"""
        if self.door_interlock_state:
            if self.door_interlock_breakabled:
                self.door_interlock_final_state = "locked_active"
                msg = "Locked (unlock enabled)"
            else:
                self.door_interlock_final_state = "locked_inactive"
                msg = "Locked (unlock disabled)"
        else:
            self.door_interlock_final_state = "unlocked"
            msg = "Unlocked"

        if not self.ics_enabled:
            self.door_interlock_final_state = "error"
            msg = "Desy ICS error"

        if not self.use_door_interlock:
            self.door_interlock_final_state = "locked_active"
            msg = "Locked (unlock enabled)"

        self.emit("doorInterlockStateChanged", self.door_interlock_final_state, msg)
        return self.door_interlock_final_state, msg

    def unlock_door_interlock(self):
        """Break Interlock (only if it is allowed by doorInterlockCanUnlock)
           It doesn't matter what we are sending in the command
           as long as it is a one char
        """
        if HWR.beamline.diffractometer is not None:
            detector_distance = HWR.beamline.detector.distance
            if HWR.beamline.diffractometer.in_plate_mode():
                if detector_distance is not None:
                    if detector_distance.get_value() < 780:
                        detector_distance.set_value(800, timeout=None)
                        while detector_distance.get_value() < 360:
                            gevent.sleep(0.01)
                        gevent.sleep(2)
            else:
                if detector_distance is not None:
                    if detector_distance.get_value() < 1099:
                        detector_distance.set_value(1100)
                        gevent.sleep(1)
            try:
                HWR.beamline.diffractometer.set_phase(
                    HWR.beamline.diffractometer.PHASE_TRANSFER, timeout=None
                )
            except Exception:
                logging.getLogger("GUI").error(
                    "Unable to set diffractometer to transfer phase"
                )

        if not self.use_door_interlock:
            logging.getLogger().info("Door interlock is disabled")
            return

        if self.door_interlock_state:
            if self.door_interlock_breakabled:
                if self.cmd_break_interlock is None:
                    self.cmd_break_interlock = self.get_command_object(
                        "cmdBreakInterlock"
                    )
                self.cmd_break_interlock()
            else:
                msg = (
                    "Door Interlock cannot be broken at the moment "
                    + "please check its status and try again."
                )
                logging.getLogger("GUI").error(msg)
        else:
            logging.getLogger("HWR").info("Door is Interlocked")

    def re_emit_values(self):
        """Updates state"""
        self.get_state()



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Gives access to the Hardware Objects contained in the Hardware Repository database

The Hardware Repository database is a set of XML files describing devices, equipments
and procedures on a beamline. Each XML file represent a Hardware Object.
The Hardware Repository module provides access to these Hardware Objects, and manages
connections to the Control Software (Spec or Taco Device Servers).
"""

from __future__ import (
    absolute_import,
    print_function,
)

import importlib
import logging
import os
import sys
import time
import traceback
import weakref
from datetime import datetime
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Optional,
    Union,
)
from warnings import warn

from ruamel.yaml import YAML

from mxcubecore import (
    BaseHardwareObjects,
    HardwareObjectFileParser,
)
from mxcubecore.dispatcher import dispatcher
from mxcubecore.protocols_config import setup_commands_channels
from mxcubecore.utils.conversion import (
    make_table,
    string_types,
)

if TYPE_CHECKING:
    from mxcubecore.BaseHardwareObjects import HardwareObject

# If you want to write out copies of the file, use typ="rt" instead
# pure=True uses yaml version 1.2, with fewer gotchas for strange type conversions
yaml = YAML(typ="safe", pure=True)
# The following are not needed for load, but define the default style.
yaml.default_flow_style = False
yaml.indent(mapping=2, sequence=4, offset=2)


__copyright__ = """ Copyright © 2010 - 2020 by MXCuBE Collaboration """
__license__ = "LGPLv3+"


_instance = None
TIMERS = []

beamline = None
BEAMLINE_CONFIG_FILE = "beamline_config.yml"


def load_from_yaml(
    configuration_file,
    role,
    yaml_export_directory: Optional[Path] = None,
    _container=None,
    _table=None,
):
    """

    Args:
        configuration_file (str):
        role (str): Role name of configured object, used as its name
        _container (ConfiguredObject): Container object for recursive loading
        _table Optional[List]: Internal, collecting summary output

    Returns:

    """
    global beamline

    column_names = ("role", "Class", "file", "Time (ms)", "Comment")
    if _table is None:
        # This is the topmopst call
        _table = []

    start_time = time.time()
    msg0 = ""
    result = None
    class_name = "None"

    # Get full path for configuration file
    if _instance is None:
        raise RuntimeError("HardwareRepository has not been initialised")
    configuration_path = _instance.find_in_repository(configuration_file)
    if configuration_path is None:
        msg0 = "File not found"

    if not msg0:
        # Load the configuration file
        with open(configuration_path, "r") as fp0:
            configuration = yaml.load(fp0)
        class_import = configuration.pop("class", None)
        if not class_import:
            if _container:
                msg0 = "No 'class' tag"
            else:
                # at top level we want to get the actual error
                raise ValueError("%s file lacks  'class' tag" % configuration_file)

    if not msg0:
        module_name, class_name = class_import.rsplit(".", 1)
        # For "a.b.c" equivalent to absolute import of "from a.b import c"
        try:
            cls = getattr(importlib.import_module(module_name), class_name)
        except Exception as ex:
            if _container:
                msg0 = "Error importing class"
                class_name = class_import
                print(
                    "Encountered Exception (continuing):\n%s" % traceback.format_exc()
                )
            else:
                # at top level we want to get the actual error
                raise

    if not msg0:
        try:
            # instantiate object
            result = cls(name=role)
            result._hwobj_container = _container
        except Exception:
            if _container:
                msg0 = "Error instantiating %s" % cls.__name__
                print(
                    "Encountered Exception (continuing):\n%s" % traceback.format_exc()
                )
            else:
                # at top level we want to get the actual error
                raise

    if _container is None:
        # We are loading the beamline object into HardwareRepository
        # and want the link to be set before _init or content loading
        beamline = result

    if not msg0:
        try:
            config = configuration.pop("configuration", {})
            # Set configuration with non-object properties.
            result._config = result.HOConfig(**config)

            # Initialise object
            result._init()
        except Exception:
            if _container:
                msg0 = "Error in %s._init()" % cls.__name__
            else:
                # at top level we want to get the actual error
                raise

    if not msg0:
        # Recursively load contained objects (of any type that the system can support)
        objects = configuration.pop("objects", {})

        setup_commands_channels(result, configuration)

        if _container is None:
            load_time = 1000 * (time.time() - start_time)
            msg1 = "Start loading contents:"
            _table.append(
                (role, class_name, configuration_file, "%.1d" % load_time, msg1)
            )
            msg0 = "Done loading contents"

        for role1, config_file in objects.items():
            fname, fext = os.path.splitext(config_file)
            if fext in (".yaml", ".yml"):
                fname = f"/{fname}"

                hwobj = load_from_yaml(
                    config_file,
                    role=role1,
                    yaml_export_directory=yaml_export_directory,
                    _container=result,
                    _table=_table,
                )
                if hwobj:
                    # only add if we successfully loaded the object
                    _instance.hardware_objects[fname] = hwobj

            elif fext == ".xml":
                msg1 = ""
                time0 = time.time()
                class_name1 = "None"
                try:
                    hwobj = _instance.get_hardware_object(fname)
                    if hwobj is None:
                        msg1 = "No object loaded"
                    else:
                        class_name1 = hwobj.__class__.__name__
                        _attach_xml_objects(yaml_export_directory, result, hwobj, role1)
                except Exception as ex:
                    msg1 = "Loading error (%s)" % str(ex)
                load_time = 1000 * (time.time() - time0)
                _table.append(
                    (role1, class_name1, config_file, "%.1d" % load_time, msg1)
                )
    if not msg0:
        if _container:
            if not hasattr(_container, role):
                warn(
                    f"load_from_yaml Class {_container.__class__.__name__} has no attribute {role}",
                    stacklevel=2,
                )
            _container._hwobj_by_role[role] = result
        try:
            # Initialise object
            result.init()
        except Exception:
            if _container:
                msg0 = "Error in %s.init()" % cls.__name__
            else:
                # at top level we want to get the actual error
                raise

    load_time = 1000 * (time.time() - start_time)
    _table.append((role, class_name, configuration_file, "%.1d" % load_time, msg0))

    if _container is None:
        print(make_table(column_names, _table))
    elif yaml_export_directory and result:
        _export_draft_config_file(yaml_export_directory, result)

    return result


def _export_draft_config_file(dest_dir: Path, hwobj):
    def write_yaml(data, file_name: str):
        dest_dir.mkdir(parents=True, exist_ok=True)
        yaml.dump(data, Path(dest_dir, file_name))

    result = {
        "class": "%s.%s" % (hwobj.__class__.__module__, hwobj.__class__.__name__),
    }
    objects_by_role = hwobj.objects_by_role
    if objects_by_role:
        objects = result["objects"] = {}
        for role, obj in objects_by_role.items():
            try:
                objects[role] = "%s.yaml" % obj.id
            except Exception:
                logging.getLogger("HWR").exception("")

    config = result["configuration"] = {}
    for tag, val in hwobj.config.model_dump().items():
        if tag not in objects_by_role:
            config[tag] = val  # noqa: PERF403

    write_yaml(result, "%s.yaml" % hwobj.id)


def _attach_xml_objects(yaml_export_directory: Optional[Path], container, hwobj, role):
    """Recursively attach XML-configured object to container as role

    NBNB guard against duplicate objects"""

    hwobj._hwobj_container = container
    hwobj._name = role
    container._hwobj_by_role[role] = hwobj

    objects_by_role = hwobj._objects_by_role
    for role2, hwobj2 in objects_by_role.items():
        _attach_xml_objects(yaml_export_directory, hwobj, hwobj2, role2)

    if yaml_export_directory and hwobj:
        _export_draft_config_file(yaml_export_directory, hwobj)


def _convert_xml_property(hwobj):
    """Convert complex xml-configured object"""
    result = {}
    result.update(hwobj.get_properties())
    for tag in hwobj._objects_names():
        # NB this does NOT allow having HardwareObjects inside complex properties
        objs = list(hwobj._get_objects(tag))
        result[tag] = [_convert_xml_property(obj) for obj in objs]
    return result


def _create_config_for_xml_hwobj(hwobj: BaseHardwareObjects.HardwareObjectNode):
    """
    Populate hwobj._config attribute for an HWOBJ loaded with XML configure file.

    This allows to access HWOBJ configuration uniformly for both YAML and XML
    configured objects, using its 'config' attribute.
    """
    hwobj._config = hwobj.HOConfig(**hwobj.get_properties())

    objects_by_role = hwobj._objects_by_role
    for tag in hwobj._objects_names():
        if tag not in objects_by_role:
            # Complex object, not contained hwobj
            objs = [_convert_xml_property(obj) for obj in hwobj._get_objects(tag)]
            if len(objs) == 1:
                setattr(hwobj.config, tag, objs[0])
            else:
                setattr(hwobj.config, tag, objs)


def add_hardware_objects_dirs(ho_dirs):
    """Adds directories with xml/yaml config files

    Args:
        ho_dirs ([type]): [description]
    """
    if isinstance(ho_dirs, list):
        new_ho_dirs = list(filter(os.path.isdir, list(map(os.path.abspath, ho_dirs))))

        for new_ho_dir in reversed(new_ho_dirs):
            if new_ho_dir not in sys.path:
                sys.path.insert(0, new_ho_dir)


def set_user_file_directory(user_file_directory):
    """Sets user file directory.

    Args:
        user_file_directory (str): absolute path to user file directory
    """
    BaseHardwareObjects.HardwareObjectNode.set_user_file_directory(user_file_directory)


def init_hardware_repository(
    configuration_path: str,
    yaml_export_directory: Optional[Path] = None,
):
    """Initialise hardware repository - must be run at program start

    Args:
        configuration_path (str): PATHSEP-separated string of directories
        giving configuration file lookup path
        yaml_export_directory: if specified, loaded hardware objects configuration
        will be written to this directory, as YAML files

    Returns:

    """
    global _instance
    global beamline

    if _instance is not None or beamline is not None:
        raise RuntimeError(
            "init_hardware_repository called on already initialised repository"
        )
    if not configuration_path:
        logging.getLogger("HWR").error(
            "Unable to initialize hardware repository. No configuration path passed."
        )
        return

    # If configuration_path is a string of combined paths, split it up
    lookup_path = [
        os.path.abspath(os.path.expanduser(x))
        for x in configuration_path.split(os.path.pathsep)
    ]
    lookup_path = [x for x in lookup_path if os.path.exists(x)]
    if lookup_path:
        configuration_path = lookup_path

    logging.getLogger("HWR").info("Hardware repository: %s", configuration_path)
    _instance = __HardwareRepositoryClient(configuration_path)
    _instance.connect()
    beamline = load_from_yaml(
        BEAMLINE_CONFIG_FILE,
        role="beamline",
        yaml_export_directory=yaml_export_directory,
    )


def uninit_hardware_repository():
    global _instance, beamline
    _instance = None
    beamline = None


def get_hardware_repository():
    """
    Get the HardwareRepository (singleton) instance,

    Returns:
        HardwareRepository: The Singleton instance of HardwareRepository
                            (in reality __HardwareRepositoryClient)
    """

    if _instance is None:
        raise RuntimeError("The HardwareRepository has not been initialised")

    return _instance


class __HardwareRepositoryClient:
    """Hardware Repository class

    Warning -- should not be instantiated directly ;
    call the module's level get_hardware_repository() function instead
    """

    def __init__(self, server_address):
        """Constructor

        server_address needs to be the HWR server address (host:port) or
        a list of paths where to find XML files locally (when server is not in use)
        """
        self.server_address = server_address
        self.required_hardware_objects = {}
        self.xml_source = {}
        self.__connected = False
        self.server = None
        self.hwobj_info_list = []
        self.invalid_hardware_objects = None
        self.hardware_objects = None

    def connect(self):
        if self.__connected:
            return
        try:
            self.invalid_hardware_objects = set()
            self.hardware_objects = weakref.WeakValueDictionary()

            self.server = None
        finally:
            self.__connected = True

    def find_in_repository(self, relative_path):
        """Finds absolute path of a file or directory matching relativePath
        in one of the hardwareRepository directories

        Will work for any file or directory, but intended for configuration
        files that do NOT match the standard XML file system"""

        if self.server:
            logging.getLogger("HWR").error(
                "Cannot find file in repository - server is in use"
            )
            return
        else:
            if relative_path.startswith(os.path.sep):
                relative_path = relative_path[1:]

            for xml_files_path in self.server_address:
                file_path = os.path.join(xml_files_path, relative_path)
                if os.path.exists(file_path):
                    return os.path.abspath(file_path)
            return

    def require(self, mnemonics_list):
        """Download a list of Hardware Objects in one go"""
        self.required_hardware_objects = {}

        if not self.server:
            return

        try:
            t0 = time.time()
            mnemonics = ",".join([repr(mne) for mne in mnemonics_list])
        except Exception:
            logging.getLogger("HWR").exception(
                "Could not execute 'require' on Hardware Repository server"
            )

    def _load_hardware_object(self, hwobj_name=""):
        """
        Load a Hardware Object. Do NOT use externally,
        as this will mess up object tracking, signals, etc.

        :param hwobj_name:  string name of the Hardware Object to load, e.g. /motors/m0
        :return: the loaded Hardware Object, or None if it fails
        """

        comment = ""
        class_name = ""
        hwobj_instance = None
        xml_data = ""

        for xml_files_path in self.server_address:
            file_name = (
                hwobj_name[1:] if hwobj_name.startswith(os.path.sep) else hwobj_name
            )
            file_path = os.path.join(xml_files_path, file_name) + os.path.extsep + "xml"
            if os.path.exists(file_path):
                try:
                    xml_data = open(file_path, "r").read()
                except Exception:
                    pass
                break

        start_time = datetime.now()

        if xml_data:
            try:
                hwobj_instance = self.parse_xml(xml_data, hwobj_name)
                if isinstance(hwobj_instance, string_types):
                    # We have redirection to another file
                    # Enter in dictionaries also under original names
                    result = self._load_hardware_object(hwobj_instance)
                    if hwobj_name in self.invalid_hardware_objects:
                        self.invalid_hardware_objects.remove(hwobj_name)
                    self.hardware_objects[hwobj_name] = result
                    return result
            except Exception:
                comment = "Cannot parse xml"
                logging.getLogger("HWR").exception(
                    "Cannot parse XML file for Hardware Object %s", hwobj_name
                )
            else:
                if hwobj_instance is not None:
                    self.xml_source[hwobj_name] = xml_data

                    dispatcher.send("hardwareObjectLoaded", hwobj_name, self)

                    def hardwareObjectDeleted(name=hwobj_instance.name):
                        logging.getLogger("HWR").debug(
                            "%s Hardware Object has been deleted from Hardware Repository",
                            name,
                        )
                        del self.hardware_objects[name]

                    hwobj_instance.resolve_references()

                    try:
                        hwobj_instance._add_channels_and_commands()
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "Error while adding commands and/or channels to Hardware Object %s",
                            hwobj_name,
                        )
                        comment = "Failed to add all commands and/or channels"

                    try:
                        _create_config_for_xml_hwobj(hwobj_instance)
                        hwobj_instance._init()
                        hwobj_instance.init()
                        class_name = str(hwobj_instance.__module__)
                    except Exception:
                        logging.getLogger("HWR").exception(
                            'Cannot initialize Hardware Object "%s"', hwobj_name
                        )
                        self.invalid_hardware_objects.add(hwobj_instance.name)
                        hwobj_instance = None
                        comment = "Failed to init class"
                    else:
                        if hwobj_instance.load_name in self.invalid_hardware_objects:
                            self.invalid_hardware_objects.remove(hwobj_instance.name)

                        self.hardware_objects[hwobj_instance.load_name] = hwobj_instance
                else:
                    logging.getLogger("HWR").error(
                        "Failed to load Hardware object %s", hwobj_name
                    )
                    comment = "Loading failed"
        else:
            logging.getLogger("HWR").error(
                'Cannot load Hardware Object "%s" : file not found.', hwobj_name
            )

        end_time = datetime.now()
        time_delta = end_time - start_time

        self.hwobj_info_list.append(
            (
                hwobj_name,
                class_name,
                "%d ms" % (time_delta.microseconds / 1000),
                comment,
            )
        )

        return hwobj_instance

    def discard_hardware_object(self, ho_name):
        """Remove a Hardware Object from the Hardware Repository

        Parameters :
          ho_name -- the name of the Hardware Object to remove

        Emitted signals :
          hardwareObjectDiscarded (<object name>) -- emitted when the object has been removed
        """
        try:
            del self.hardware_objects[ho_name]
        except KeyError:
            pass
        try:
            self.invalid_hardware_objects.remove(ho_name)
        except Exception:
            pass
        try:
            del self.required_hardware_objects[ho_name]
        except KeyError:
            pass

        dispatcher.send("hardwareObjectDiscarded", ho_name, self)

    def parse_xml(self, xml_string, ho_name):
        """Load a Hardware Object from its XML string representation

        Parameters :
          xml_string -- the XML string
          ho_name -- the name of the Hardware Object to load (i.e. '/motors/m0')

        Return :
          the Hardware Object, or None if it fails
        """
        try:
            hardware_obj = HardwareObjectFileParser.parse_string(xml_string, ho_name)
        except Exception:
            logging.getLogger("HWR").exception(
                "Cannot parse Hardware Repository file %s", ho_name
            )
        else:
            return hardware_obj

    def update(self, name, updatesList):
        """[summary]

        Args:
            name ([type]): [description]
            updatesList ([type]): [description]
        """
        # TODO: update without HWR server
        if self.server is not None and self.server.isSpecConnected():
            self.server.send_msg_cmd_with_return(
                'xml_multiwrite("%s", "%s")' % (name, str(updatesList))
            )
        else:
            logging.getLogger("HWR").error(
                "Cannot update Hardware Object %s : not connected to server", name
            )

    def rewrite_xml(self, name, xml):
        """[summary]

        Args:
            name ([type]): [description]
            xml ([type]): [description]
        """
        # TODO: rewrite without HWR server
        if self.server is not None and self.server.isSpecConnected():
            self.server.send_msg_cmd_with_return(
                'xml_writefile("%s", %s)' % (name, repr(xml))
            )
            self.xml_source[name] = xml
        else:
            logging.getLogger("HWR").error(
                "Cannot update Hardware Object %s : not connected to server", name
            )

    def __getitem__(self, item):
        """[summary]

        Args:
            item ([type]): [description]

        Raises:
            KeyError: [description]

        Returns:
            [type]: [description]
        """
        if item == "equipments":
            return self.get_equipments()
        elif item == "procedures":
            return self.get_procedures()
        elif item == "devices":
            return self.get_devices()
        else:
            return self.get_hardware_object(item)

        raise KeyError

    def get_equipments(self):
        """Return the list of the currently loaded Equipments Hardware Objects"""
        eq_list = []

        for ho_name in self.hardware_objects:
            if self.is_equipment(ho_name):
                eq_list.append(self.hardware_objects[ho_name])

        return eq_list

    def get_procedures(self):
        """Return the list of the currently loaded Procedures Hardware Objects"""
        result = []

        for ho_name in self.hardware_objects:
            if self.is_procedure(ho_name):
                result.append(self.hardware_objects[ho_name])

        return result

    def get_devices(self):
        """Return the list of the currently loaded Devices Hardware Objects"""
        result = []

        for ho_name in self.hardware_objects:
            if self.is_device(ho_name):
                result.append(self.hardware_objects[ho_name])

        return result

    def get_hardware_object(self, object_name: str) -> Union["HardwareObject", None]:
        """Return a Hardware Object given its name.

        If the object is not in the Hardware Repository, try to load it.

        Args:
            object_name (str): The name of the Hardware Object

        Returns:
            Union[HardwareObject, None]: The required Hardware Object
        """

        if not object_name:
            return None

        if not object_name.startswith("/"):
            object_name = "/" + object_name

        try:
            if object_name:
                if object_name in self.invalid_hardware_objects:
                    return None

                if object_name in self.hardware_objects:
                    hardware_obj = self.hardware_objects[object_name]
                else:
                    hardware_obj = self._load_hardware_object(object_name)
                return hardware_obj
        except TypeError as err:
            logging.getLogger("HWR").exception(
                "could not get Hardware Object %s", object_name
            )

    def get_equipment(self, equipment_name):
        """Return an Equipment given its name (see get_hardware_object())"""
        return self.get_hardware_object(equipment_name)

    def get_device(self, device_name):
        """Return a Device given its name (see get_hardware_object())"""
        return self.get_hardware_object(device_name)

    def get_procedure(self, procedure_name):
        """Return a Procedure given its name (see get_hardware_object())"""
        return self.get_hardware_object(procedure_name)

    def is_device(self, name):
        """Check if a Hardware Object is a Device

        Parameters :
          name -- name of the Hardware Object to test

        Return :
          True if the Hardware Object is a Device, False otherwise
        """
        try:
            return isinstance(self.hardware_objects[name], BaseHardwareObjects.Device)
        except Exception:
            return False

    def is_procedure(self, name):
        """Check if a Hardware Object is a Procedure

        Parameters :
          name -- name of the Hardware Object to test

        Return :
          True if the Hardware Object is a Procedure, False otherwise
        """
        try:
            return isinstance(
                self.hardware_objects[name], BaseHardwareObjects.Procedure
            )
        except Exception:
            return False

    def is_equipment(self, name):
        """Check if a Hardware Object is an Equipment

        Parameters :
          name -- name of the Hardware Object to test

        Return :
          True if the Hardware Object is an Equipment, False otherwise
        """
        try:
            return isinstance(
                self.hardware_objects[name], BaseHardwareObjects.Equipment
            )
        except Exception:
            return False

    def has_hardware_object(self, name):
        """Check if the Hardware Repository contains an object

        Parameters :
          name -- name of the Hardware Object

        Return :
          True if HardwareObject is loaded in the Hardware Repository, False otherwise
        """
        return name in self.hardware_objects

    def get_info(self, name):
        """Return a dictionary with information about the specified Hardware Object

        Parameters :
          name -- name of the Hardware Object

        Return :
          a dictionary containing information about the Hardware Object
        """
        try:
            hardware_obj = self.hardware_objects[name]
        except KeyError:
            return {}
        else:
            hardware_obj_class = hardware_obj.__class__.__name__

            d = {
                "class": hardware_obj_class,
                "python module": sys.modules[hardware_obj.__module__].__file__,
            }

            if hasattr(hardware_obj, "is_ready"):
                d["is ready ?"] = str(hardware_obj.is_ready())

            if hasattr(hardware_obj, "get_commands"):
                # hardware object is a command container
                d["commands"] = {}

                for cmd in hardware_obj.get_commands():
                    if cmd.__class__.__name__ == "SpecCommand":
                        d["commands"][cmd.userName()] = {
                            "type": "spec",
                            "version": "%s:%s"
                            % (
                                cmd.connection.host,
                                cmd.connection.port or cmd.connection.scanname,
                            ),
                            "connected ?": cmd.isSpecConnected() and "yes" or "no",
                            "macro or function": str(cmd.command),
                        }
                    elif cmd.__class__.__name__ == "TacoCommand":
                        dd = {"type": "taco", "device": cmd.device_name}

                        try:
                            dd["imported ?"] = "yes" if cmd.device.imported else "no"
                        except Exception:
                            dd["imported ?"] = "no, invalid Taco device"

                        dd["device method"] = str(cmd.command)

                        d["commands"][cmd.userName()] = dd
                    elif cmd.__class__.__name__ == "TangoCommand":
                        d["commands"][cmd.userName()] = {
                            "type": "tango",
                            "device": cmd.device_name,
                            "imported ?": (
                                "no, invalid Tango device"
                                if cmd.device is None
                                else "yes"
                            ),
                            "device method": str(cmd.command),
                        }

                d["channels"] = {}

                for chan in hardware_obj.get_channels():
                    if chan.__class__.__name__ == "SpecChannel":
                        d["channels"][chan.userName()] = {
                            "type": "spec",
                            "version": "%s:%s"
                            % (
                                chan.connection.host,
                                chan.connection.port or chan.connection.scanname,
                            ),
                            "connected ?": chan.isSpecConnected() and "yes" or "no",
                            "variable": str(chan.varName),
                        }
                    elif chan.__class__.__name__ == "TangoChannel":
                        d["channels"][chan.userName()] = {
                            "type": "tango",
                            "device": chan.device_name,
                            "imported ?": chan.device is not None
                            and "yes"
                            or "no, invalid Tango device or attribute name",
                            "attribute": str(chan.attribute_name),
                        }

            if "SpecMotorA" in [
                klass.__name__ for klass in hardware_obj.__class__.__bases__
            ]:
                d["spec version"] = hardware_obj.specVersion
                d["motor mnemonic"] = hardware_obj.specName
                try:
                    d["connected ?"] = (
                        "yes" if hardware_obj.connection.isSpecConnected() else "no"
                    )
                except Exception:
                    d["connected ?"] = "no"

            if isinstance(hardware_obj, BaseHardwareObjects.DeviceContainer):
                d["children"] = {}

                for ho in hardware_obj.get_devices():
                    try:
                        d["children"][ho.load_name] = self.get_info(ho.load_name)
                    except Exception:
                        continue

            return d

    def end_polling(self):
        """Stop all pollers

        Warning : should not be used directly (finalization purposes only)
        """
        return

    def close(self):
        """'close' the Hardware Repository

        Discards all Hardware Objects
        """
        self.end_polling()

        self.hardware_objects = weakref.WeakValueDictionary()

    def timerEvent(self, t_ev):
        try:
            global TIMERS

            func_ref = TIMERS[t_ev.timerId()]
            func = func_ref()

            if func is None:
                # self.killTimer(t_ev.timerId())
                del TIMERS[t_ev.timerId()]
            else:
                try:
                    func()
                except Exception:
                    logging.getLogger("HWR").exception(
                        "an error occured while calling timer function"
                    )
        except Exception:
            logging.getLogger("HWR").exception("an error occured inside the timerEvent")

    def print_report(self):
        longest_cols = [
            (max([len(str(row[i])) for row in self.hwobj_info_list]) + 3)
            for i in range(len(self.hwobj_info_list[0]))
        ]
        row_format = "| ".join(
            ["{:<" + str(longest_col) + "}" for longest_col in longest_cols]
        )

        print("+", "=" * sum(longest_cols), "+")
        print("| %s" % row_format.format(*("xml", "Class", "Load time", "Comment")))
        print("+", "=" * sum(longest_cols), "+")

        for row in sorted(self.hwobj_info_list):
            print("| %s" % row_format.format(*row))
        print("+", "=" * sum(longest_cols), "+")

    def reload_hardware_objects(self):
        """
        Reloads all modified modules.
        Package reimport is used to detect modified modules.
        Hardware objects that correspond to these modules:
        1. are disconnected from gui
        2. imported in this module with __import__ and reimport is called
        3. connected back to the gui channels
        """
        # NOTE
        # reimport is supported for python 2.x and not by python 3.x
        # if needed a similar package for 3x could be used. In this case
        # code depends on a platform: platform.python_version()[0] > 2 ...

        # NB reload_hardware_objects does NOT work with beamline_opbject
        # and other yaml configs

        import reimport

        modified_modules = reimport.modified()
        for hwr_obj in self.hardware_objects:
            for item in modified_modules:
                if self.hardware_objects[hwr_obj].__module__ == item:
                    try:
                        connections = self.hardware_objects[hwr_obj].connect_dict
                        for sender in connections:
                            self.hardware_objects[hwr_obj].disconnect(
                                sender,
                                connections[sender]["signal"],
                                connections[sender]["slot"],
                            )
                        logging.getLogger("HWR").debug(
                            "HardwareRepository: %s disconnected from GUI", item
                        )
                        self.hardware_objects[hwr_obj].clear_gevent()
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "HardwareRepository: Unable to disconnect hwobj %s", item
                        )
                        continue

                    try:
                        __import__(item, globals(), locals(), [], -1)
                        reimport.reimport(item)
                        logging.getLogger("HWR").debug(
                            "HardwareRepository: %s reloaded", item
                        )
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "HardwareRepository: Unable to reload module %s", item
                        )

                    try:
                        for sender in connections:
                            self.hardware_objects[hwr_obj].connect(
                                sender,
                                connections[sender]["signal"],
                                connections[sender]["slot"],
                            )
                        logging.getLogger("HWR").debug(
                            "HardwareRepository: %s connected to GUI", item
                        )
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "HardwareRepository: Unable to connect hwobj %s", item
                        )
                    try:
                        self.hardware_objects[hwr_obj].init()
                        self.hardware_objects[hwr_obj].re_emit_values()
                        logging.getLogger("HWR").debug(
                            "HardwareRepository: %s initialized and updated", item
                        )
                    except Exception:
                        logging.getLogger("HWR").exception(
                            "HardwareRepository: Unable to initialize hwobj %s", item
                        )



# encoding: utf-8
#
#  Project: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
Harvester Maintenance.
"""

import logging

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject


class HarvesterMaintenance(HardwareObject):
    __TYPE__ = "HarvesterMaintenance"

    """
    Actual implementation of the Harvester MAINTENANCE,
    COMMANDS, Actions and Calibration procedure, and Centering.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def init(self):
        self._harvester = self.get_object_by_role("harvester")

    def _do_trash(self):
        """
        Trash sample

        :returns: None
        """
        return self._harvester.trash_sample()

    def _transfer_sample(self):
        """
        Transfer sample

        :returns: None
        """
        return self._harvester.transfer_sample()

    def _load_plate(self, args):
        """
        Load Plate

        :returns: None
        :args: str
        """
        return self._harvester.load_plate(plate_id=args)

    def _set_room_temperature_mode(self, args):
        """
        Set Harvester temperature mode

        :returns: None
        :args: boolean
        """
        return self._harvester.set_room_temperature_mode(args)

    def _do_abort(self):
        """
        Abort current command

        :returns: None
        """
        return self._harvester.abort()

    def _do_park(self):
        """
        Abort and Park (Homing)

        :returns: None
        """
        self._harvester.do_abort()
        return self._harvester.home()

    def _update_global_state(self):
        """
        update global state
        :returns: True
        """
        state_dict, cmd_state, message = self.get_global_state()
        self.emit("globalStateChanged", (state_dict, cmd_state, message))
        return True

    def get_global_state(self):
        """
        update global state
        :returns: True
        """
        state = self._harvester.get_state()
        # ready = self._harvester._is_device_busy()
        running = state in ["RUNNING", "running"]
        plate_ID = self._harvester.get_plate_id()

        state_dict = {
            "running": running,
            "state": state,
            "plate_barecode": plate_ID or "Null",
        }

        cmd_state = {
            "transfer": True,
            "trash": True,
            "park": True,
            "abort": True,
        }

        message = ""

        return state_dict, cmd_state, message

    def get_cmd_info(self):
        """return information about existing commands for this object"""
        """ [cmd_id, cmd_display_name, nb_args, cmd_category, description ] """

        cmd_list = [
            [
                "Actions",
                [
                    ["transfer", "Transfer sample", "Actions", None],
                    ["trash", "Trash sample", "Actions", None],
                    ["park", "Park", "Actions", None],
                    ["abort", "Abort", "Actions", None],
                ],
            ],
        ]

        return cmd_list

    def send_command(self, cmd_name, args=None):
        if cmd_name in ["park"]:
            self._do_park()
        if cmd_name == "trash":
            self._do_trash()
        if cmd_name == "transfer":
            self._transfer_sample()
        if cmd_name == "abort":
            self._do_abort()
        if cmd_name == "loadPlateWithBarcode":
            self._load_plate(args)
        if cmd_name == "set_room_temperature_mode":
            self._set_room_temperature_mode(args)
        return True

    def calibrate_pin(self) -> bool:
        """
            Pin Calibration Procedure
            In other for the Centring procedure to work on a Harvested Pin
            a Calibration Procedure need to be execute

        Return (bool): whether the calibration procedure goes to end (True)
        or had and exception (False)
        """

        self._harvester.load_calibrated_pin()
        self._harvester._wait_sample_transfer_ready(None)
        print("waiting 40 seconds before mount")
        # For some reason the Harvester return READY too soon
        # approximately 40 Second sooner
        gevent.sleep(40)
        sample_mount_device = HWR.beamline.sample_changer
        mount_current_sample = sample_mount_device.load_a_pin_for_calibration()

        if mount_current_sample:
            try:
                md = HWR.beamline.diffractometer
                md._wait_ready()

                sample_drift_x = float(self._harvester.get_last_sample_drift_offset_x())
                sample_drift_y = float(self._harvester.get_last_sample_drift_offset_y())
                sample_drift_z = float(
                    -self._harvester.get_last_sample_drift_offset_z()
                )

                motor_pos_dict = {
                    "kappa": float(
                        md["HacentringReferencePosition"].get_property("kappa_ref")
                    ),
                    "kappa_phi": float(
                        md["HacentringReferencePosition"].get_property("phi_ref")
                    ),
                    "phi": float(
                        md["HacentringReferencePosition"].get_property("omega_ref")
                    ),
                    "phiy": md.phiyMotor.get_value() + sample_drift_x,
                }

                md.move_motors(motor_pos_dict)
                md._wait_ready()
                md.centringFocus.set_value_relative(sample_drift_z, None)
                md.centringVertical.set_value_relative(sample_drift_y, None)

                md.save_current_motor_position()
                self._harvester.set_calibration_state(True)

                logging.getLogger("user_level_log").info(
                    "Pin Calibration Step 1 Succeed"
                )
                logging.getLogger("user_level_log").info(
                    "User Need to Perform an  3 click centring"
                )
                return True
            except Exception:
                logging.getLogger("user_level_log").exception("Pin Calibration Failed")
                return False
        else:
            logging.getLogger("user_level_log").error("Pin Calibration Failed")
            logging.getLogger("user_level_log").error(
                "Sample Changer could not mount Pin"
            )
            return False

    def validate_calibration(self) -> bool:
        """
        finish Calibration Procedure step 2
        after user ran a 3 click centering
        Return (bool): whether the step 2 of calibration procedure
        goes to end (True) or had and exception (False)
        """
        try:
            md = HWR.beamline.diffractometer

            motor_pos_dict = {
                "focus": md.focusMotor.get_value(),
                "phiy": md.phiyMotor.get_value(),
                "phiz": md.phizMotor.get_value(),
                "centring_focus": md.centringFocus.get_value(),
                "centring_vertical": md.centringVertical.get_value(),
            }

            saved_position = md.saved_motor_position
            # find offset position based on old and new motor position
            new_motor_offset = {
                "focus": motor_pos_dict["focus"] - saved_position["focus"],
                "phiy": motor_pos_dict["phiy"] - saved_position["phiy"],
                "phiz": motor_pos_dict["phiz"] - saved_position["phiz"],
                "centring_focus": (
                    motor_pos_dict["centring_focus"] - saved_position["centring_focus"]
                ),
                "centring_vertical": (
                    motor_pos_dict["centring_vertical"]
                    - saved_position["centring_vertical"]
                ),
            }

            calibrated_motor_offset = {
                "focus": new_motor_offset["focus"] + new_motor_offset["centring_focus"],
                "phiy": new_motor_offset["phiy"],
                "phiz": (
                    new_motor_offset["phiz"] + new_motor_offset["centring_vertical"]
                ),
            }

            # we store the motor offset in the Harvester, to be used for sample centering
            self._harvester.store_calibrated_pin(
                calibrated_motor_offset["focus"],
                calibrated_motor_offset["phiy"],
                calibrated_motor_offset["phiz"],
            )

            self._harvester.set_calibration_state(False)
        except Exception:
            logging.getLogger("user_level_log").exception(
                "Pin Calibration / validation Failed"
            )
            return False

        return True



from pydantic.v1 import (
    BaseModel,
    Field,
)


class ExporterNStateConfigModel(BaseModel):
    _class: str = Field("", alias="@class")
    defualt_value: bool = Field(False)
    exporter_address: str
    use_hwstate: bool = Field(True)
    username: str
    value_channel_name: str
    values: str


class Motor(BaseModel):
    _class: str = Field("", alias="@class")
    exporter_address: str
    username: str
    motor_name: str
